package dbrown;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * Represents a chef in a restaurant who prepares orders until the restaurant
 * closes. After finishing an order the chef notifies the waiter that the order
 * is ready.
 *
 * @author David B. Brown from Bruce Eckel
 * @version 2015-12-02
 */
class Chef implements Runnable {
    // Constants.
    public static final int FREE = 0;
    public static final int COOKING = 1;
    public static final int DONE = 2;
    private static final int OFFSET = 1000;
    private static final int DELAY = 500;

    private Order order = null;

    private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);

    private final Restaurant restaurant;

    private int status = Chef.FREE;

    /**
     * The chef constructor.
     *
     * @param r
     *            The restaurant the chef works in.
     */
    public Chef(final Restaurant r) {
	this.restaurant = r;
    }

    /**
     * Assigns listeners to this object.
     *
     * @param listener
     */
    public void addPropertyChangeListener(
	    final PropertyChangeListener listener) {
	this.pcs.addPropertyChangeListener(listener);
    }

    /**
     * Returns the order the chef is currently cooking.
     *
     * @return Returns the order.
     */
    public Order getOrder() {
	return this.order;
    }

    /**
     * Returns the chef's current status.
     *
     * @return Returns the status.
     */
    public int getStatus() {
	return this.status;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.lang.Runnable#run()
     */
    @Override
    public synchronized void run() {
	try {
	    // Wait for the first order.
	    this.wait();

	    while (!this.restaurant.getClosed()) {
		// An order arrives.
		this.status = Chef.COOKING;
		this.pcs.firePropertyChange(null, null, this.status);
		// The chef cooks for a while.
		Thread.sleep((long) Math.random() * DELAY + OFFSET);

		// Inform the restaurant the chef is done.
		synchronized (this.restaurant) {
		    this.status = Chef.FREE;
		    this.pcs.firePropertyChange(null, null, this.status);
		    this.order = null;
		    this.restaurant.notify();
		}
		// Wait until restaurant notifies us of the next order.
		this.wait();
	    }
	} catch (final InterruptedException e) {
	    // Nothing done.
	}
	// End of shift.
	this.status = Chef.DONE;
	this.pcs.firePropertyChange(null, null, this.status);
    }

    /**
     * Assigns an order to the chef.
     *
     * @param order
     *            The order to set.
     */
    public void setOrder(final Order order) {
	this.order = order;
    }

}