package dbrown;

import java.awt.Dimension;
import java.awt.Point;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * Main class for the Restaurant simulation. Spawns one Chef, multiple waiters,
 * and their viewers, and passes one order at a time between the Chef and
 * Waiters.
 *
 * @author David B. Brown from Bruce Eckel
 * @version 2015-12-02
 */
public class Restaurant implements Runnable {
    // Constants.
    public static final int WIDTH = 200;
    public static final int HEIGHT = 100;
    // Attributes to handle screen placement.
    public static final Dimension SIZE = new Dimension(WIDTH, HEIGHT);
    private final Point location = new Point(400, 200);

    private static final int OFFSET = 500;
    private static final int DELAY = 100;

    /**
     * Number of waiters in restaurant.
     */
    public static final int WAITERS = 3;

    /**
     * <code>threadPool</code> provides an unbounded pool of execution threads.
     */
    private final ExecutorService waiterThreadPool = Executors
	    .newFixedThreadPool(Restaurant.WAITERS);
    private final ExecutorService chefThread = Executors
	    .newSingleThreadExecutor();

    /**
     * The Restaurant's Chef object.
     */
    private final Chef chef = new Chef(this);

    /**
     * Tracks whether the restaurant is closed or not.
     */
    private boolean closed = false;

    /**
     * Restaurant owns and creates the original Orders, and controls their
     * sharing between chef and waiter.
     */
    private Order order = null;

    /**
     * The Waiters.
     */
    private final Waiter[] waiters = new Waiter[Restaurant.WAITERS];

    /**
     * Allows views to listen to generic changes in the model.
     */
    private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);

    /**
     * Restaurant constructor. Creates the Chef and Waiter objects and their
     * respective viewers. Starts the Restaurant thread.
     */
    public Restaurant() {

	final RestaurantView restaurantView = new RestaurantView(
		Restaurant.this, Restaurant.this.location);
	restaurantView.setVisible(true);
	this.location.setLocation(this.location.x + WIDTH,
		this.location.y + HEIGHT);

	// Define the chef and its viewer.
	final ChefView chefView = new ChefView(Restaurant.this.chef,
		Restaurant.this.location, Restaurant.SIZE);
	this.chefThread.execute(this.chef);
	chefView.setVisible(true);

	// Define the waiters and their viewers.
	for (int i = 0; i < Restaurant.WAITERS; i++) {
	    this.nextLocation();
	    final Waiter waiter = new Waiter(this);
	    final WaiterView waiterView = new WaiterView(waiter,
		    Restaurant.this.location, Restaurant.SIZE);
	    this.waiters[i] = waiter;
	    this.waiterThreadPool.execute(this.waiters[i]);
	    waiterView.setVisible(true);
	}
    }

    /**
     * Assigns listeners to this object.
     *
     * @param listener
     */
    public void addPropertyChangeListener(
	    final PropertyChangeListener listener) {
	this.pcs.addPropertyChangeListener(listener);
    }

    /**
     * Returns whether the restaurant is closed or not.
     *
     * @return whether the restaurant is closed.
     */
    public boolean getClosed() {
	return this.closed;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.lang.Runnable#run()
     */
    @Override
    public synchronized void run() {
	do {
	    try {
		// Generate a new order.
		this.order = new Order();

		if (this.order.isLastOrder()) {
		    this.pcs.firePropertyChange(null, null,
			    "Last order, closing soon.");
		}
		this.pcs.firePropertyChange(null, null,
			"New order: " + this.order.getNumber());

		// Give the order to the chef.
		synchronized (this.chef) {
		    this.chef.setOrder(this.order);
		    this.chef.notify();
		}
		this.pcs.firePropertyChange(null, null, "Passed order to chef");
		// Wait for the chef to finish cooking.
		this.wait();
		// Pass the order to the first available waiter.
		int i = (int) Math.random() * Restaurant.WAITERS;

		// Poll the waiters to see who is free.
		while (this.waiters[i].getStatus() == Waiter.SERVING) {
		    i = (i + 1) % Restaurant.WAITERS;
		}

		// Give the order to the waiter.
		synchronized (this.waiters[i]) {
		    this.waiters[i].setOrder(this.order);
		    this.waiters[i].notify();
		    this.pcs.firePropertyChange(null, null, "Order "
			    + this.order.getNumber() + " Given to Waiter " + i);
		}
		Thread.sleep((long) Math.random() * Restaurant.DELAY
			+ Restaurant.OFFSET);
	    } catch (final InterruptedException e) {
		// Nothing done.
	    }
	} while (!this.order.isLastOrder());

	this.closed = true;
	// Tell the chef the restaurant is closed.
	synchronized (this.chef) {
	    this.chef.notify();
	}
	this.chefThread.shutdown();
	this.waiterThreadPool.shutdown();
	this.pcs.firePropertyChange(null, null, "Restaurant Closed");
    }

    // -------------------------------------------------------------------------------
    /**
     * Determines the location of the next waiter/chef frame. Lays out frames
     * from top-to-bottom.
     */
    private void nextLocation() {
	final int x = this.location.x;
	final int y = this.location.y + HEIGHT;
	this.location.setLocation(x, y);
    }

}
