package dbrown;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

/**
 * Simulates a simple waiter who can handle a single order at a time, and waits
 * until an order is available before doing anything.
 *
 * @author David B. Brown from Bruce Eckel
 * @version 2015-12-02
 */
public class Waiter implements Runnable {

    // Constants for Waiter status.
    public static final int FREE = 0;
    public static final int SERVING = 1;
    public static final int DONE = 2;
    private static final int OFFSET = 5000;
    private static final int DELAY = 5000;

    /**
     * Tracks the total number of waiters generated.
     */
    private static int count = 0;

    private Order order = null;

    private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);

    private final Restaurant restaurant;

    private int status = Waiter.FREE;

    /**
     * The number of a particular waiter. This number can be accessed using the
     * getter <code>getNumber</code>, but cannot be set.
     */
    private final int number = Waiter.count++;

    /**
     * The <code>Waiter</code> constructor.
     *
     * @param r
     *            The restaurant the waiter works at.
     */
    public Waiter(final Restaurant r) {
	this.restaurant = r;
    }

    /**
     * Assigns listeners to this object.
     *
     * @param listener
     */
    public void addPropertyChangeListener(
	    final PropertyChangeListener listener) {
	this.pcs.addPropertyChangeListener(listener);
    }

    /**
     * Returns the ID number of the current waiter.
     *
     * @return the current ID number.
     */
    public int getNumber() {
	return this.number;
    }

    /**
     * Returns the current order held by the waiter.
     *
     * @return Returns the order.
     */
    public Order getOrder() {
	return this.order;
    }

    /**
     * Returns the current status of the waiter.
     *
     * @return Returns the status.
     */
    public int getStatus() {
	return this.status;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.lang.Runnable#run()
     */
    @Override
    public synchronized void run() {
	// The waiter works until the restaurant closes.
	while (!this.restaurant.getClosed()) {
	    try {
		// The waiter can accept another order.
		this.status = Waiter.FREE;
		this.pcs.firePropertyChange(null, null, this.status);
		// Have the waiter wait until given an order.
		this.wait();
		// Serve the order.
		this.status = Waiter.SERVING;
		this.pcs.firePropertyChange(null, null, this.status);
		// The waiter must take time to serve the order.
		Thread.sleep(
			(long) Math.random() * Waiter.DELAY + Waiter.OFFSET);
		this.order = null;
	    } catch (final InterruptedException e) {
		// nothing to do.
	    }
	}
	// End of shift.
	this.status = Waiter.DONE;
	this.pcs.firePropertyChange(null, null, this.status);
    }

    /**
     * Assigns an order to the waiter.
     *
     * @param order
     *            The order to set.
     */
    public void setOrder(final Order order) {
	this.order = order;
    }

}